#define _BSD_SOURCE
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "crcf.h"

static int hex2dec(char c)
{
	if (c >= '0' && c <= '9')
		return c - '0';
	if (c >= 'a' && c <= 'f')
		return c - 'a' + 10;
	if (c >= 'A' && c <= 'F')
		return c - 'A' + 10;
	return 0;
}

char *crcf_str_encode(char *in)
{
	size_t leni, leno;
	char *out;
	size_t offseti, offseto;
	
	leni = strlen(in);
	leno = leni;
	out = malloc(leno+1);
	for (offseti = 0, offseto = 0; offseti < leni; offseti++, offseto++) {
		if (isalnum(in[offseti])) {
			out[offseto] = in[offseti];
		} else {
			leno+=3;
			out = realloc(out, leno + 1);
			sprintf(out+offseto, "%%%02x", in[offseti]);
			offseto+=2;
		}
	}
	out[offseto] = 0;
	
	return out;
}

char *crcf_str_decode(char *in)
{
	char *out;
	size_t offseti, offseto;
	size_t len;
	
	len = strlen(in);
	out = malloc(len+1);
	
	for (offseti = 0, offseto = 0; offseti < len; offseti++, offseto++) {
		if (in[offseti] == '%') {
			out[offseto] = 
			    (hex2dec(in[offseti+1]) << 4) |
			    hex2dec(in[offseti+2]);
			offseti+=2;
		} else {
			out[offseto] = in[offseti];
		}
	}
	out[offseto] = 0;
	
	return out;
}

static size_t crcf_db_size = 0;
static struct crcf_id **crcf_db = NULL;

struct crcf_id *crcf_db_iterate(struct crcf_id *prev)
{
	size_t i;
	
	if (!crcf_db_size)
		return NULL;
	
	if (!prev) {
		return crcf_db[0];
	}
	
	for (i = 0; i < crcf_db_size - 1; i++) {
		if (crcf_db[i] == prev)
			return crcf_db[i+1];
	}

	return NULL;
}

struct crcf_id *crcf_db_search_id(char *id)
{
	size_t i;
	
	for (i = 0; i < crcf_db_size; i++) {
		if (!strcmp(crcf_db[i]->id, id))
			return crcf_db[i];
	}

	return NULL;
}

struct crcf_id *crcf_db_search_id_actor(char *id, enum crcf_actor actor)
{
	size_t i;
	
	for (i = 0; i < crcf_db_size; i++) {
		if (!strcmp(crcf_db[i]->id, id) &&
		    crcf_db[i]->actor == actor)
			return crcf_db[i];
	}

	return NULL;
}

struct crcf_protocol *crcf_db_search_protocol(int bus, int address)
{
	struct crcf_id *entry = NULL;
	
	while ((entry = crcf_db_iterate(entry))) {
		struct crcf_protocol *protocol;
		
		if (entry->actor != CRCF_ACTOR_PROTOCOL)
			continue;
		protocol = (struct crcf_protocol *)entry;
		
		if (protocol->bus != bus)
			continue;
		if (protocol->address != address)
			continue;
		
		return protocol;
	}


	return NULL;
}


int crcf_db_add(struct crcf_id *entry)
{
	struct crcf_id **db;
	
	db = realloc(crcf_db, sizeof(struct crcf_id *) * (crcf_db_size + 1));
	if (!db)
		return -1;
	
	crcf_db = db;
	crcf_db[crcf_db_size] = entry;
	crcf_db_size++;
	
	return 0;
}

static int crcf_id_base = 2;

int crcf_id_create(struct crcf_id *id, char *idstr, enum crcf_actor actor)
{
	memset(id->id, 0, CRCF_ID_SIZE);
	if (idstr) {
		strcpy(id->id, idstr);
	} else {
		sprintf(id->id, "%d", crcf_id_base++);
	}
	id->actor = actor;

	return 0;
}

bool crcf_id_broadcast(char *id)
{
	if (!strcmp(id, "0"))
		return true;
	if (!strcmp(id, CRCF_ID_BROADCAST))
		return true;
	return false;
}

struct crcf_loco *crcf_loco_create(char *id)
{
	struct crcf_loco *loco = NULL;
	
	if (id) {
		/* Check if it already exists */
		loco = (struct crcf_loco *)crcf_db_search_id(id);
	}
	
	if (!loco) {
		loco = calloc(sizeof(struct crcf_loco), 1);
		
		crcf_id_create(&loco->id, id, CRCF_ACTOR_LOCO);
		
		if (!loco)
			return NULL;
		
		crcf_db_add(&loco->id);
	} else {
		/* Check existing type */
		if (loco->id.actor != CRCF_ACTOR_LOCO)
			return NULL;
	}
	
	return loco;
}

int crcf_loco_name_set(struct crcf_loco *loco, char *name)
{
	strncpy(loco->name, name, CRCF_NAME_LEN);

	return 0;
}

char *crcf_loco_name_get(struct crcf_loco *loco)
{
	return loco->name;
}

int crcf_loco_decoder_set(struct crcf_loco *loco, struct crcf_decoder *decoder)
{
	memcpy(&loco->decoder, &decoder->id, sizeof(struct crcf_id));

	return 0;
}

struct crcf_decoder *crcf_decoder_create(char *id)
{
	struct crcf_decoder *decoder = NULL;
	
	if (id) {
		/* Check if it already exists */
		decoder = (struct crcf_decoder *)crcf_db_search_id(id);
	}
	
	if (!decoder) {
		decoder = calloc(sizeof(struct crcf_decoder), 1);
		
		crcf_id_create(&decoder->id, id, CRCF_ACTOR_DECODER);
		
		if (!decoder)
			return NULL;
		
		crcf_db_add(&decoder->id);
	} else {
		/* Check existing type */
		if (decoder->id.actor != CRCF_ACTOR_DECODER)
			return NULL;
	}
	
	return decoder;
}

int crcf_decoder_protocol_set(struct crcf_decoder *decoder, struct crcf_protocol *protocol)
{
	memcpy(&decoder->protocol, &protocol->id, sizeof(struct crcf_id));

	return 0;
}

int crcf_decoder_loco_set(struct crcf_decoder *decoder, struct crcf_loco *loco)
{
	memcpy(&decoder->loco, &loco->id, sizeof(struct crcf_id));
	
	return 0;
}

struct crcf_loco *crcf_decoder_loco_get(struct crcf_decoder *decoder)
{
	struct crcf_loco *loco;

	loco = (struct crcf_loco *)
	    crcf_db_search_id_actor(decoder->loco.id, CRCF_ACTOR_LOCO);
	
	if (!loco)
		return NULL;

	return loco;
}

struct crcf_protocol *crcf_protocol_create(char *id)
{
	struct crcf_protocol *protocol = NULL;
	
	if (id) {
		/* Check if it already exists */
		protocol = (struct crcf_protocol *)crcf_db_search_id(id);
	}
	
	if (!protocol) {
		protocol = calloc(sizeof(struct crcf_protocol), 1);
		
		crcf_id_create(&protocol->id, id, CRCF_ACTOR_PROTOCOL);
		
		if (!protocol)
			return NULL;
		
		crcf_db_add(&protocol->id);
	} else {
		/* Check existing type */
		if (protocol->id.actor != CRCF_ACTOR_PROTOCOL)
			return NULL;
	}
	
	return protocol;
}

struct crcf_decoder *crcf_protocol_decoder_get(struct crcf_protocol *protocol)
{
	struct crcf_decoder *decoder;

	decoder = (struct crcf_decoder *)
	    crcf_db_search_id_actor(protocol->decoder.id, CRCF_ACTOR_DECODER);
	
	if (!decoder)
		return NULL;

	return decoder;
}

int crcf_protocol_address_set(struct crcf_protocol *protocol, int address)
{
	protocol->address = address;
	
	return 0;
}

int crcf_protocol_bus_set(struct crcf_protocol *protocol, int bus)
{
	protocol->bus = bus;
	
	return 0;
}

int crcf_protocol_speedsteps_set(struct crcf_protocol *protocol, int speedsteps)
{
	protocol->speedsteps = speedsteps;
	
	return 0;
}

int crcf_protocol_decoder_set(struct crcf_protocol *protocol, struct crcf_decoder *decoder)
{
	memcpy(&protocol->decoder, &decoder->id, sizeof(struct crcf_id));
	
	return 0;
}

struct crcf_function *crcf_db_protocol_search_function(struct crcf_protocol *protocol, int number)
{
	struct crcf_id *entry = NULL;
	
	while ((entry = crcf_db_iterate(entry))) {
		struct crcf_function *function;
		
		if (entry->actor != CRCF_ACTOR_FUNCTION)
			continue;
		
		function = (struct crcf_function*)entry;
		
		if (memcmp(function->protocol.id, protocol->id.id, CRCF_ID_SIZE))
			continue;
		
		if (function->nr == number)
			return function;
	}

	return NULL;
}

struct crcf_function *crcf_function_create(char *id)
{
	struct crcf_function *function = NULL;
	
	if (id) {
		/* Check if it already exists */
		function = (struct crcf_function *)crcf_db_search_id(id);
	}
	
	if (!function) {
		function = calloc(sizeof(struct crcf_function), 1);
		
		crcf_id_create(&function->id, id, CRCF_ACTOR_FUNCTION);
		
		if (!function)
			return NULL;
		
		crcf_db_add(&function->id);
	} else {
		/* Check existing type */
		if (function->id.actor != CRCF_ACTOR_FUNCTION)
			return NULL;
	}
	
	return function;
}

struct crcf_protocol *crcf_function_protocol_get(struct crcf_function *function)
{
	struct crcf_protocol *protocol;

	protocol = (struct crcf_protocol *)
	    crcf_db_search_id_actor(function->protocol.id, CRCF_ACTOR_PROTOCOL);
	
	if (!protocol)
		return NULL;

	return protocol;
}

struct crcf_function *crcf_db_iterate_protocol_function(struct crcf_protocol *protocol, struct crcf_function *prev)
{
	struct crcf_id *entry = NULL;
	
	if (prev) {
		while ((entry = crcf_db_iterate(entry))) {
			if (!memcmp(entry->id, prev->id.id, CRCF_ID_SIZE)) {
				break;
			}
		}
	}
	while ((entry = crcf_db_iterate(entry))) {
		struct crcf_function *function;
		
		if (entry->actor != CRCF_ACTOR_FUNCTION)
			continue;
		
		function = (struct crcf_function *)entry;
		
		if (!memcmp(function->protocol.id, protocol->id.id, CRCF_ID_SIZE)) {
			return function;
		}
	}
	return NULL;
}

int crcf_db_search_protocol_functioncount(struct crcf_protocol *protocol)
{
	struct crcf_function *entry = NULL;
	int count = 0;
	
	while ((entry = crcf_db_iterate_protocol_function(protocol, entry)))
		count++;
	
	return count;
}

int crcf_function_number_set(struct crcf_function *function, int nr)
{
	function->nr = nr;
	
	return 0;
}

int crcf_function_mode_set(struct crcf_function *function, int pushbutton)
{
	function->pushbutton = pushbutton;
	
	return 0;
}

int crcf_function_number_get(struct crcf_function *function)
{
	return function->nr;
}

int crcf_function_mode_get(struct crcf_function *function)
{
	return function->pushbutton;
}

int crcf_function_protocol_set(struct crcf_function *function, struct crcf_protocol *protocol)
{
	memcpy(&function->protocol, &protocol->id, sizeof(struct crcf_id));
	
	return 0;
}

