#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <linux/joystick.h>

#include "srcp.h"
#include "udev_search.h"
#include "display.h"
#include "crcf.h"

static bool send_update = false;

struct srcp *srcp_cmd;
struct srcp *srcp_info;
struct srcp_loc_dcc *loc = NULL;

int fd_display = -1;

bool joystick_locked = false;

static void locodb_trigger(void)
{
	srcp_message_set(srcp_cmd, 0, srcp_session_id_get(srcp_info), "CRCF",
	    "LOCODB " CRCF_ID_BROADCAST " LIST LOCO");
}

static char *loc_name(int bus, int address)
{
	struct crcf_protocol *protocol;
	struct crcf_decoder *decoder;
	struct crcf_loco *loco;
	
	protocol = crcf_db_search_protocol(bus, address);
	if (!protocol)
		return NULL;
	
	decoder =  crcf_protocol_decoder_get(protocol);
	if (!decoder)
		return NULL;

	loco =  crcf_decoder_loco_get(decoder);
	if (!loco)
		return NULL;
	
	return crcf_loco_name_get(loco);
}

void check_lock(void)
{
	if (!loc)
		return;
		
	if (loc->lock && loc->lock == srcp_session_id_get(srcp_cmd)) {
		joystick_locked = true;
	} else {
		joystick_locked = false;
	}
}


void term_lock(void)
{
	if (loc && joystick_locked) {
		srcp_lock_loc_dcc_term(srcp_cmd, loc);
	}
}

void event_lock(void)
{
	if (joystick_locked)
		term_lock();
	else
		srcp_lock_loc_dcc_set(srcp_cmd, loc);
}

static bool console_name = false;

static void srcp_console_status(struct srcp_loc_dcc *loc)
{
	int i;
	char *name;
	
	if (!loc)
		return;
	
	name = loc_name(loc->bus, loc->address);
	if (!name) {
		name = "";
		console_name = false;
		locodb_trigger();
	} else
		console_name = true;
	
	display_name(name);
	
	display_address(loc->address);

	printf("%s %03d/%03d ",
	    loc->drivemode == SRCP_LOC_DCC_DRIVEMODE_FORWARD ? " Forward" : "Backward",
	    loc->V, loc->V_max);

	display_speed(loc->V, loc->V_max, 
	    loc->drivemode == SRCP_LOC_DCC_DRIVEMODE_FORWARD);

	display_functions(loc->nrofdecoderfunctions);

	for (i = 0; i < loc->nrofdecoderfunctions; i++) {
		printf("%c", loc->functions[i] ? 'o' : '.');
		display_function(i, loc->functions[i]);
	}
	printf("\n");

	display_locked(fd_display, loc->lock, !joystick_locked);

	display_update(fd_display);
}


void check_console_name(void)
{
	char *name;

	if (!loc)
		return;
	
	if (console_name)
		return;
	
	name = loc_name(loc->bus, loc->address);
	if (name)
		srcp_console_status(loc);
}


void event_power_set(bool state)
{
	srcp_power_set(srcp_cmd, 1, state);
}

void event_loc_next(void)
{
	term_lock();

	loc = srcp_loc_dcc_list_next(srcp_info, loc);
	if (!loc)
		loc = srcp_loc_dcc_list_next(srcp_info, loc);
	if (loc) {
		printf("Switched to loc %d\n",
		    loc->address);
		check_lock();
		srcp_console_status(loc);
	}
}

void event_loc_prev(void)
{
	term_lock();

	loc = srcp_loc_dcc_list_prev(srcp_info, loc);
	if (!loc)
		loc = srcp_loc_dcc_list_prev(srcp_info, loc);
	if (loc) {
		printf("Switched to loc %d\n",
		    loc->address);
		check_lock();
		srcp_console_status(loc);
	}
}
void event_function_set(int funcnr, bool value)
{
	struct crcf_protocol *protocol;
	struct crcf_function *function = NULL;
	int mode = 0;
					
	protocol = crcf_db_search_protocol(loc->bus, loc->address);
	if (protocol) {
		function = crcf_db_protocol_search_function(protocol, funcnr);
	}
	if (function) {
		mode = crcf_function_mode_get(function);
	}
					
	if (mode == 0) {
		if (value) {
			loc->functions[funcnr] = 
			    !loc->functions[funcnr];
		}
	} else {
		loc->functions[funcnr] = value;
	}

}

static void loc_dcc_init_hook(struct srcp *srcp, struct srcp_loc_dcc *loc_new)
{
	printf("New loc: %d\n", loc_new->address);
	if (!loc) {
		loc = loc_new;
		printf("Using this loc\n");
	} else
		srcp_loc_dcc_put(loc_new);
}

static void loc_dcc_hook(struct srcp *srcp, struct srcp_loc_dcc *loc_info)
{
	if (loc_info == loc)
		srcp_console_status(loc_info);
	srcp_loc_dcc_put(loc_info);
}

static void lock_loc_dcc_hook(struct srcp *srcp, struct srcp_loc_dcc *loc_info)
{
	if (loc_info == loc) {
		check_lock();
		srcp_console_status(loc_info);
	}
	srcp_loc_dcc_put(loc_info);
}

static void power_hook(struct srcp *srcp, bool power)
{
	printf("Power: %d\n", power);
	display_power(fd_display, power);
}

struct controller {
	int fd;

	int axis_speed;

	int axis_select;
	bool axis_select_np; // becomes true when below half positive
	bool axis_select_nn; // becomes true when above half negative

	int button_direction;
	int button_shift_10;
	int button_shift_20;
	int button_esc;
	
	int shift;
	bool esc;
};

void joystick_event(struct controller *controller)
{
	struct js_event event;
	int ret;
	int max_event = 10;
	
	do {
		ret = read (controller->fd, &event, sizeof(struct js_event));
		if (ret == sizeof(struct js_event)) {
			if (event.type & JS_EVENT_BUTTON) {
				if (event.number == controller->button_esc) {
					controller->esc = event.value;
					printf("ESC: %d\n", controller->esc);
					continue;
				} else if (controller->esc) {
					if (event.number == 1 && event.value) {
						if (event.value)
							event_power_set(false);
					}
					if (event.number == 2 && event.value) {
						if (event.value)
							event_power_set(true);
					}
					if (event.number == 3 && event.value) {
						event_lock();
					}
					continue;
				}
				if (!loc || !joystick_locked) {
					continue;
				}
				if (event.number == controller->button_direction) {
					if (event.value) {
						loc->drivemode = 
						    loc->drivemode == SRCP_LOC_DCC_DRIVEMODE_FORWARD ?
						    SRCP_LOC_DCC_DRIVEMODE_BACKWARD :
						    SRCP_LOC_DCC_DRIVEMODE_FORWARD;
					}
				} else if (event.number == controller->button_shift_10) {
					if (event.value) {
						controller->shift = 10;
					} else {
						controller->shift = 0;
					}
				} else if (event.number == controller->button_shift_20) {
					if (event.value) {
						controller->shift = 20;
					} else {
						controller->shift = 0;
					}
				} else if (event.number + controller->shift < loc->nrofdecoderfunctions) {
					int funcnr;
					
					funcnr = event.number + controller->shift;
					
					event_function_set(funcnr, event.value);
				}
				send_update = true;
			}
			if (event.type & JS_EVENT_AXIS && loc) {
				if (event.number == controller->axis_speed && joystick_locked) {
					int value = event.value;
					
					if (value > 0)
						value = 0;
					/* forward is negative on USB joysticks */
					value = -value;

					value *= loc->V_max;
					value += 16768;
					value /= 32768;
					if (value > loc->V_max)
						value = loc->V_max;
					
					if (value != loc->V) {
						loc->V = value;
						send_update = true;
					}
				} else if (event.number == controller->axis_select) {
					if (event.value == 32767 &&
					    controller->axis_select_np) {
					    	event_loc_next();
						controller->axis_select_np = false;
					} else if (event.value == -32767 &&
					    controller->axis_select_nn) {
					    	event_loc_prev();
						controller->axis_select_nn = false;
					}
					if (event.value > -16384)
						controller->axis_select_nn = true;
					if (event.value < 16384)
						controller->axis_select_np = true;
				}
			}
		}
	} while (ret == sizeof(struct js_event) && --max_event);
}

static int handler_locodb_info_loco(int reply_to, char *actor_id, char *attribute_value)
{
	char *reply;
	
	crcf_loco_create(attribute_value);
	
	asprintf(&reply, "LOCO %s GET NAME", attribute_value);
	
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	
	free(reply);
	
	asprintf(&reply, "LOCO %s LIST DECODER", attribute_value);
	
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	
	free(reply);

	return 0;
}

static int handler_loco_info_name(int reply_to, char *actor_id, char *attribute_value)
{
	char *name = crcf_str_decode(attribute_value);
	struct crcf_loco *loco;
	
	printf("name: %s\n", name);
	
	loco = crcf_loco_create(actor_id);
	if (loco) {
		crcf_loco_name_set(loco, name);
	}
	
	free(name);
	
	return 0;
}

static int handler_loco_info_decoder(int reply_to, char *actor_id, char *attribute_value)
{
	char *reply;
	struct crcf_loco *loco;
	struct crcf_decoder *decoder;
	
	decoder = crcf_decoder_create(attribute_value);
	loco = crcf_loco_create(actor_id);
	if (decoder && loco) {
		crcf_decoder_loco_set(decoder, loco);
		crcf_loco_decoder_set(loco, decoder);
	}
	
	asprintf(&reply, "DECODER %s LIST PROTOCOL", attribute_value);
	
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	
	free(reply);
	
	return 0;
}

static int handler_decoder_info_protocol(int reply_to, char *actor_id, char *attribute_value)
{
	char *reply;
	struct crcf_decoder *decoder;
	struct crcf_protocol *protocol;
	
	decoder = crcf_decoder_create(actor_id);
	protocol = crcf_protocol_create(attribute_value);
	if (decoder && protocol) {
		crcf_decoder_protocol_set(decoder, protocol);
		crcf_protocol_decoder_set(protocol, decoder);
	}
	
	asprintf(&reply, "PROTOCOL %s GET BUS", attribute_value);
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	free(reply);
	
	asprintf(&reply, "PROTOCOL %s GET ADDRESS", attribute_value);
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	free(reply);

	asprintf(&reply, "PROTOCOL %s LIST FUNCTION", attribute_value);
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	free(reply);
	
	return 0;
}

static int handler_protocol_info_bus(int reply_to, char *actor_id, char *attribute_value)
{
	struct crcf_protocol *protocol;
	
	protocol = crcf_protocol_create(actor_id);
	if (protocol) {
		int bus = atoi(attribute_value);

		crcf_protocol_bus_set(protocol, bus);
	}

	check_console_name();
	
	return 0;
}

static int handler_protocol_info_address(int reply_to, char *actor_id, char *attribute_value)
{
	struct crcf_protocol *protocol;
	
	protocol = crcf_protocol_create(actor_id);
	if (protocol)
		crcf_protocol_address_set(protocol, atoi(attribute_value));
	
	check_console_name();
	
	return 0;
}

static int handler_protocol_info_function(int reply_to, char *actor_id, char *attribute_value)
{
	char *reply;
	struct crcf_protocol *protocol;
	struct crcf_function *function;
	
	protocol = crcf_protocol_create(actor_id);
	function = crcf_function_create(attribute_value);
	if (function && protocol) {
		crcf_function_protocol_set(function, protocol);
	}
	
	asprintf(&reply, "FUNCTION %s GET NUMBER", attribute_value);
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	free(reply);
	
	asprintf(&reply, "FUNCTION %s GET MODE", attribute_value);
	srcp_message_set(srcp_cmd, reply_to, srcp_session_id_get(srcp_info), "CRCF",
	    reply);
	free(reply);

	return 0;
}

static int handler_function_info_number(int reply_to, char *actor_id, char *attribute_value)
{
	struct crcf_function *function;
	
	function = crcf_function_create(actor_id);
	if (function)
		crcf_function_number_set(function, atoi(attribute_value));
	
	return 0;
}

static int handler_function_info_mode(int reply_to, char *actor_id, char *attribute_value)
{
	struct crcf_function *function;
	
	function = crcf_function_create(actor_id);
	if (function)
		crcf_function_mode_set(function, atoi(attribute_value));
	
	return 0;
}

struct crcf_message {
	char *actor;
	char *method;
	char *attribute;
	int (*handle)(int reply_to, char *actor_id, char *attribute_value);
};

static struct crcf_message handlers[] = {
	{ "LOCODB",   "INFO", "LOCO",     handler_locodb_info_loco },
	{ "LOCO",     "INFO", "NAME",     handler_loco_info_name },
	{ "LOCO",     "INFO", "DECODER",  handler_loco_info_decoder },
	{ "DECODER",  "INFO", "PROTOCOL", handler_decoder_info_protocol },
	{ "PROTOCOL", "INFO", "BUS",      handler_protocol_info_bus },
	{ "PROTOCOL", "INFO", "ADDRESS",  handler_protocol_info_address },
	{ "PROTOCOL", "INFO", "FUNCTION", handler_protocol_info_function },
	{ "FUNCTION", "INFO", "NUMBER",   handler_function_info_number },
	{ "FUNCTION", "INFO", "MODE",     handler_function_info_mode },
};

static void message_hook(struct srcp *srcp, int send_to, int reply_to, char *type, char *msg)
{
	if (!strcmp(type, "CRCF")) {
		char *actor, *actor_id, *method, *attribute, *attribute_value;
		int i;
		
		sscanf(msg, "%ms %ms %ms %ms %ms", 
		    &actor, &actor_id, &method, &attribute, & attribute_value);
		
		for (i = 0; i < sizeof(handlers)/sizeof(struct crcf_message); i++) {
			if (!strcmp(actor, handlers[i].actor) &&
			    !strcmp(method, handlers[i].method) &&
			    !strcmp(attribute, handlers[i].attribute)) {
				handlers[i].handle(reply_to, actor_id, attribute_value);
				break;
			}
		}
		
		free(actor);
		free(actor_id);
		free(method);
		free(attribute);
		free(attribute_value);
	}
}

int main(int argc, char **argv)
{
	struct termios oldtio,newtio;
	char *joystick_device;
	int fd_joystick;
	int fd_srcp_info;
	int fd_sdl;
	char nr_axes;
	char nr_buttons;
	int fd_max = 0;
	struct controller controller = { };
	
	/* Some sane values for a 'normal' joystick */
	controller.axis_speed = 1;
	controller.axis_select = 0;
	controller.button_direction = 1;
	controller.button_shift_10 = -1;
	controller.button_shift_20 = -1;
	controller.button_esc = -1;
	controller.axis_select_np = true;
	controller.axis_select_nn = true;

	if (argc < 3) {
		printf("Usage %s [host] [joystick device] <speed axisnr>\n", 
		    argv[0]);
		return -1;
	}
	if (argc > 3) {
		controller.axis_speed = atoi(argv[3]);
	}
	
	joystick_device = argv[2];
	printf("Joystick device: %s\n", joystick_device);
	fd_joystick = open(joystick_device, O_RDONLY);
	if (fd_joystick < 0) {
		printf("Could not open device %s\n", joystick_device);
		printf("Going to simulate a joystick\n");
	} else {
		ioctl(fd_joystick, FIONBIO, &(int){1});
		ioctl(fd_joystick, JSIOCGAXES, &nr_axes);
		ioctl(fd_joystick, JSIOCGBUTTONS, &nr_buttons);
	}
	printf("%s: %d axes, %d buttons\n", joystick_device, nr_axes, nr_buttons);
	printf("Using axis %d for speed\n", controller.axis_speed);

	if (nr_axes == 2 && nr_buttons == 14) {
		struct js_corr corr[ABS_CNT];
		char *hidraw_device;
		
		ioctl(fd_joystick, JSIOCGCORR, corr);
		printf("type: %d %d %d %d %d\n", corr[0].type, 
		    corr[0].coef[0], corr[0].coef[1], corr[0].coef[2], corr[0].coef[3]);
		printf("type: %d %d %d %d %d\n", corr[1].type, 
		    corr[1].coef[0], corr[1].coef[1], corr[1].coef[2], corr[1].coef[3]);
		corr[0].type = JS_CORR_BROKEN;
		corr[1].type = JS_CORR_BROKEN;
		corr[1].coef[0] = 0;
		corr[1].coef[1] = 0;
		corr[1].coef[2] = (1<<14) * 128;
		corr[1].coef[3] = (1<<14) * 128;
		ioctl(fd_joystick, JSIOCSCORR, corr);
		ioctl(fd_joystick, JSIOCGCORR, corr);
		
		printf("Found USB controller\n");
		
		controller.button_direction = 12;
		controller.button_shift_10 = 10;
		controller.button_shift_20 = 11;
		controller.button_esc = 13;

		hidraw_device = udev_search_js2hidraw(joystick_device);

		if (hidraw_device) {
			fd_display = open(hidraw_device, O_RDWR);
		}
	} else {
		printf("Unknown joystick type, keep 'simple' button mapping.\n");
	}

	fd_sdl = display_init(fd_display);
	display_locked(fd_display, joystick_locked, false);
	
	printf("Using axis %d for speed\n", controller.axis_speed);
	printf("Opening connection to %s\n", argv[1]);
	
	srcp_cmd = srcp_create(argv[1], "4303");
	if (!srcp_cmd) {
		perror("srcp_create()");
		return 1;
	}

	srcp_connectionmode_set(srcp_cmd, SRCP_CONNECTIONMODE_COMMAND);
	srcp_go(srcp_cmd);
	
	printf("Opening info connection to %s\n", argv[1]);
	srcp_info = srcp_create(argv[1], "4303");
	if (!srcp_info) {
		perror("srcp_create()");
		return 1;
	}
	srcp_connectionmode_set(srcp_info, SRCP_CONNECTIONMODE_INFO);

	srcp_hook_loc_dcc_init(srcp_info, loc_dcc_init_hook);
	srcp_hook_loc_dcc(srcp_info, loc_dcc_hook);
	srcp_hook_power(srcp_info, power_hook);
	srcp_hook_lock_loc_dcc(srcp_info, lock_loc_dcc_hook);

	srcp_go(srcp_info);
	fd_srcp_info = srcp_fd_get(srcp_info);
	printf("fd_srcp_info: %d\n", fd_srcp_info);

	srcp_hook_message(srcp_info, message_hook);

	tcgetattr(0, &oldtio);
	newtio = oldtio;
	newtio.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP
                | INLCR | IGNCR | ICRNL | IXON);
	newtio.c_lflag &= ~(ECHO | ICANON | ISIG | IEXTEN);
	newtio.c_cflag &= ~(CSIZE | PARENB);
	newtio.c_cflag |= CS8;
	tcsetattr(0, TCSANOW, &newtio);

	fd_max = 1;
	if (fd_joystick >= 0)
		fd_max = fd_joystick + 1;
	if (fd_srcp_info >= fd_max)
		fd_max = fd_srcp_info + 1;
	if (fd_sdl >= 0 &&  fd_sdl >= fd_max)
		fd_max = fd_srcp_info + 1;
		
	controller.fd = fd_joystick;
	
	while (1) {
		char in;
		fd_set read_set;
		struct timeval to;

		if (loc && send_update) {
			if (joystick_locked)
				srcp_loc_dcc_set(srcp_cmd, loc);
			send_update = false;
		}
		
		FD_ZERO(&read_set);
		FD_SET(0, &read_set);
		FD_SET(fd_srcp_info, &read_set);

		if (fd_joystick >= 0)
			FD_SET(fd_joystick, &read_set);
		if (fd_sdl >= 0)
			FD_SET(fd_sdl, &read_set);
		
		to.tv_sec = 1;
		to.tv_usec = 0;
		
		select(fd_max, &read_set, NULL, NULL, &to);

		if (fd_joystick >= 0 && FD_ISSET(fd_joystick, &read_set)) {
			joystick_event(&controller);
		} else	if (FD_ISSET(fd_srcp_info, &read_set)) {
			if (srcp_info_handle(srcp_info) == SRCP_RESPONSE_ERROR_NO_DATA) {
				goto out_grey;
			}
		}

		if (FD_ISSET(0, &read_set)) {
			read(0, &in, 1);
		} else if (fd_sdl >= 0 && FD_ISSET(fd_sdl, &read_set)) {
			read(fd_sdl, &in, 1);
		} else {
			display_update(fd_display);
			continue;
		}
		
		if (in == 'q') {
			goto out_grey;
		}

		if (!loc) {
			continue;
		}
		switch (in) {
			case '.':
				loc->V++;
				if (loc->V > loc->V_max)
					loc->V = loc->V_max;
				break;
			case ',':
				loc->V--;
				if (loc->V < 0)
					loc->V = 0;
				break;
			case 'd':
				loc->drivemode = 
				    loc->drivemode == SRCP_LOC_DCC_DRIVEMODE_FORWARD ?
				    SRCP_LOC_DCC_DRIVEMODE_BACKWARD :
				    SRCP_LOC_DCC_DRIVEMODE_FORWARD;

			case 'h':
				loc->V = 0;
				break;
			case ']':
				event_loc_next();
				break;
			case '[':
				event_loc_prev();
				break;
			case 'a':
				event_power_set(false);
				break;
			case 's':
				event_power_set(true);
				break;
			case 'l':
				event_lock();
				break;
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
				event_function_set(in - '0', true);
				break;
			default:
				break;
		}
		send_update = true;

	}

out_grey:
	display_grey(fd_display);
	tcsetattr(0, TCSANOW, &oldtio);

	return 0;
}
