/*
	srcp.c
		
	SRCP client implementation

	Copyright Jeroen Vreeken (jeroen@vreeken.net), 2012, 2013, 2014

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.

 */
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#include "srcp.h"

struct srcp {
	int sock;
	FILE *fdsock;

	bool power;
	int session_id;
	struct srcp_loc_dcc *loc_dcc_list;
	struct srcp_accessory_dcc *accessory_dcc_list;
	
	void (*all_hook)(struct srcp *srcp, char *message);

	void (*loc_dcc_init_hook)(struct srcp *srcp, struct srcp_loc_dcc *loc);
	void (*loc_dcc_hook)(struct srcp *srcp, struct srcp_loc_dcc *loc);
	void (*lock_loc_dcc_hook)(struct srcp *srcp, struct srcp_loc_dcc *loc);

	void (*accessory_dcc_init_hook)(struct srcp *srcp, struct srcp_accessory_dcc *accessory);
	void (*accessory_dcc_hook)(struct srcp *srcp, struct srcp_accessory_dcc *accessory);
	void (*lock_accessory_dcc_hook)(struct srcp *srcp, struct srcp_accessory_dcc *accessory);

	void(*power_hook)(struct srcp *srcp, bool power);

	void (*message_hook)(struct srcp *srcp, int send_to, int reply_to, char *type, char *msg);
};

struct srcp *srcp_create(char *host, char *port)
{
	struct addrinfo *result;
	struct addrinfo *entry;
	struct addrinfo hints = { 0 };
	struct srcp *srcp;
	int error, i;
	int sock = -1;
	char readbuffer[100];
	
	hints.ai_family = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	
	error = getaddrinfo(host, port, &hints, &result);
	if (error) {
		fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(error));
		return NULL;
	}
	
	for (entry = result; entry; entry = entry->ai_next) {
		sock = socket(entry->ai_family, entry->ai_socktype,
		    entry->ai_protocol);
		if (sock >= 0) {
			if (connect(sock, entry->ai_addr, entry->ai_addrlen)) {
				close(sock);
				sock = -1;
			} else {
//				i = 1;
//				setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &i, sizeof (int));
				i = 1;
				setsockopt (sock, IPPROTO_TCP, TCP_NODELAY, &i, sizeof (int));
				break;
			}
		}
	}
	freeaddrinfo(result);
	
	if (sock < 0)
		return NULL;
	

	srcp = calloc(sizeof(struct srcp), 1);
	
	if (!srcp)
		goto undo_socket;
	
	srcp->sock = sock;
	srcp->fdsock = fdopen(sock, "w+");
	if (!srcp->fdsock)
		goto undo_malloc;
	setvbuf(srcp->fdsock, NULL, _IONBF, 0);
	
	if (fgets(readbuffer, sizeof(readbuffer), srcp->fdsock)) {
	} else {
		goto undo_malloc;
	}

	return srcp;

undo_malloc:
	free(srcp);
undo_socket:
	close(sock);
	return NULL;
}

int srcp_fd_get(struct srcp *srcp)
{
	return srcp->sock;
}

int srcp_session_id_get(struct srcp *srcp)
{
	return srcp->session_id;
}

int srcp_hook_all(struct srcp *srcp,
    void(*all_hook)(struct srcp *srcp, char *message))
{
	srcp->all_hook = all_hook;

	return 0;
}

int srcp_hook_loc_dcc_init(struct srcp *srcp, 
    void(*loc_dcc_init_hook)(struct srcp *srcp, struct srcp_loc_dcc *loc))
{
	srcp->loc_dcc_init_hook = loc_dcc_init_hook;

	return 0;
}

int srcp_hook_loc_dcc(struct srcp *srcp, 
    void(*loc_dcc_hook)(struct srcp *srcp, struct srcp_loc_dcc *loc))
{
	srcp->loc_dcc_hook = loc_dcc_hook;

	return 0;
}

int srcp_hook_lock_loc_dcc(struct srcp *srcp,
    void(*lock_loc_dcc_hook)(struct srcp *srcp, struct srcp_loc_dcc *loc))
{
	srcp->lock_loc_dcc_hook = lock_loc_dcc_hook;
	
	return 0;
}

int srcp_hook_accessory_dcc_init(struct srcp *srcp, 
    void(*accessory_dcc_init_hook)(struct srcp *srcp, struct srcp_accessory_dcc *accessory))
{
	srcp->accessory_dcc_init_hook = accessory_dcc_init_hook;

	return 0;
}

int srcp_hook_accessory_dcc(struct srcp *srcp, 
    void(*accessory_dcc_hook)(struct srcp *srcp, struct srcp_accessory_dcc *accessory))
{
	srcp->accessory_dcc_hook = accessory_dcc_hook;

	return 0;
}

int srcp_connectionmode_set(struct srcp *srcp, enum srcp_connectionmode mode)
{
	char readbuffer[100];

	fprintf(srcp->fdsock, "SET CONNECTIONMODE SRCP %s\n",
	    mode == SRCP_CONNECTIONMODE_COMMAND ? "COMMAND" : "INFO");
	fflush(srcp->fdsock);
	if (fgets(readbuffer, sizeof(readbuffer), srcp->fdsock))
		if (srcp->all_hook)
			srcp->all_hook(srcp, readbuffer);

	return 0;
}

void srcp_loc_dcc_list_add(struct srcp *srcp, struct srcp_loc_dcc *loc)
{
	struct srcp_loc_dcc *entry, **entryp;
	
	for (entry = srcp->loc_dcc_list; entry; entry = entry->next)
		if (entry == loc)
			return;
	
	loc->refcount++;
	for (entryp = &srcp->loc_dcc_list; *entryp; entryp = &(*entryp)->next) {
		if ((*entryp)->address > loc->address)
			break;
	}
	loc->next = *entryp;
	*entryp = loc;

	if (srcp->loc_dcc_init_hook) {
		loc->refcount++;
		srcp->loc_dcc_init_hook(srcp, loc);
	}
}

void srcp_accessory_dcc_list_add(struct srcp *srcp, struct srcp_accessory_dcc *accessory)
{
	struct srcp_accessory_dcc *entry, **entryp;
	
	for (entry = srcp->accessory_dcc_list; entry; entry = entry->next)
		if (entry == accessory)
			return;
	
	accessory->refcount++;
	for (entryp = &srcp->accessory_dcc_list; *entryp; entryp = &(*entryp)->next) {
		if ((*entryp)->address > accessory->address)
			break;
	}
	accessory->next = *entryp;
	*entryp = accessory;

	if (srcp->accessory_dcc_init_hook) {
		accessory->refcount++;
		srcp->accessory_dcc_init_hook(srcp, accessory);
	}
}

static struct srcp_loc_dcc *srcp_loc_dcc_list_find(struct srcp *srcp, int address) {
	struct srcp_loc_dcc *loc;
	
	for (loc = srcp->loc_dcc_list; loc; loc = loc->next) {
		if (loc->address == address) {
			loc->refcount++;
			return loc;
		}
	}
	return NULL;
}

static struct srcp_accessory_dcc *srcp_accessory_dcc_list_find(struct srcp *srcp, int address) {
	struct srcp_accessory_dcc *accessory;
	
	for (accessory = srcp->accessory_dcc_list; accessory; accessory = accessory->next) {
		if (accessory->address == address) {
			accessory->refcount++;
			return accessory;
		}
	}
	return NULL;
}

struct srcp_loc_dcc *srcp_loc_dcc_list_next(struct srcp *srcp, struct srcp_loc_dcc *loc)
{
	struct srcp_loc_dcc *retloc;

	if (!loc) {
		retloc = srcp->loc_dcc_list;
	} else {
		retloc = loc->next;
		srcp_loc_dcc_put(loc);
	}
	if (retloc) {
		srcp_loc_dcc_take(retloc);
	}
	return retloc;
}

struct srcp_loc_dcc *srcp_loc_dcc_list_prev(struct srcp *srcp, struct srcp_loc_dcc *loc)
{
	struct srcp_loc_dcc *retloc;

	if (loc == srcp->loc_dcc_list) {
		retloc = NULL;
	} else {
		for (retloc = srcp->loc_dcc_list; retloc; retloc = retloc->next)
			if (retloc->next == loc)
				break;
	}
	if (loc) {
		srcp_loc_dcc_put(loc);
	}
	if (retloc) {
		srcp_loc_dcc_take(retloc);
	}
	return retloc;
}

void srcp_loc_dcc_put(struct srcp_loc_dcc *loc)
{
	loc->refcount--;
}

void srcp_loc_dcc_take(struct srcp_loc_dcc *loc)
{
	loc->refcount++;
}

void srcp_accessory_dcc_put(struct srcp_accessory_dcc *accessory)
{
	accessory->refcount--;
}

void srcp_accessory_dcc_take(struct srcp_accessory_dcc *accessory)
{
	accessory->refcount++;
}

static int srcp_response_get_go(struct srcp *srcp)
{
	char readbuffer[1024];
	int retcode = -2;
	
	if (fgets(readbuffer, sizeof(readbuffer), srcp->fdsock)) {
		float rep_time;
		int session_id;
		
		if (srcp->all_hook)
			srcp->all_hook(srcp, readbuffer);
		
		sscanf(readbuffer, "%f %d OK GO %d",
		    &rep_time, &retcode, &session_id);
		
		if (retcode == SRCP_RESPONSE_OK) {
			/* nothing to do */
			srcp->session_id = session_id;
		}
	}

	return retcode;	
}

static int handler_gl_init(struct srcp *srcp, char *buffer)
{
	struct srcp_loc_dcc *loc;
	float rep_time;
	int bus;
	int address;
	char protocol[1024];
	int protocolversion;
	int decoderspeedsteps;
	int nrofdecoderfunctions;
	int retcode;

	sscanf(buffer, "%f %d INFO %d GL %d "
	    "%s %d %d "
	    "%d",
	    &rep_time, &retcode, &bus, &address,
	    protocol, &protocolversion, &decoderspeedsteps,
	    &nrofdecoderfunctions);

	loc = calloc(sizeof(struct srcp_loc_dcc), 1);
	if (!loc)
		return -1;

	loc->bus = bus;
	loc->address = address;
	loc->protocolversion = protocolversion;
	loc->decoderspeedsteps = decoderspeedsteps;
	loc->nrofdecoderfunctions = nrofdecoderfunctions;

	srcp_loc_dcc_list_add(srcp, loc);

	return 0;
}


static int handler_gl_get(struct srcp *srcp, char *buffer)
{
	int drivemode;
	int V, V_max;
	int f0, f1, f2, f3, f4;
	int f5, f6, f7, f8;
	int f9, f10, f11, f12;
	int f13, f14, f15, f16, f17, f18, f19, f20;
	int f21, f22, f23, f24, f25, f26, f27, f28;
	int r, retcode, bus, address;
	float rep_time;
	struct srcp_loc_dcc *loc;
	
	r = sscanf(buffer, "%f %d INFO %d GL %d"
	    " %d %d %d"
	    " %d %d %d %d %d"
	    " %d %d %d %d"
	    " %d %d %d %d"
	    " %d %d %d %d %d %d %d %d"
	    " %d %d %d %d %d %d %d %d",
	    &rep_time, &retcode, &bus, &address,
	    &drivemode, &V, &V_max,
	    &f0, &f1, &f2, &f3, &f4,
	    &f5, &f6, &f7, &f8,
	    &f9, &f10, &f11, &f12, 
	    &f13, &f14, &f15, &f16, &f17, &f18, &f19, &f20, 
	    &f21, &f22, &f23, &f24, &f25, &f26, &f27, &f28 );

	loc = srcp_loc_dcc_list_find(srcp, address);

	loc->drivemode = drivemode;
	loc->V = V;
	loc->V_max = V_max;
	loc->functions[0] = f0;
	loc->functions[1] = f1;
	loc->functions[2] = f2;
	loc->functions[3] = f3;
	loc->functions[4] = f4;
	loc->functions[5] = f5;
	loc->functions[6] = f6;
	loc->functions[7] = f7;
	loc->functions[8] = f8;
	loc->functions[9] = f9;
	loc->functions[10] = f10;
	loc->functions[11] = f11;
	loc->functions[12] = f12;
	loc->functions[13] = f13;
	loc->functions[14] = f14;
	loc->functions[15] = f15;
	loc->functions[16] = f16;
	loc->functions[17] = f17;
	loc->functions[18] = f18;
	loc->functions[19] = f19;
	loc->functions[20] = f20;
	loc->functions[21] = f21;
	loc->functions[22] = f22;
	loc->functions[23] = f23;
	loc->functions[24] = f24;
	loc->functions[25] = f25;
	loc->functions[26] = f26;
	loc->functions[27] = f27;
	loc->functions[28] = f28;
	loc->nrofdecoderfunctions = r - 7;

	if (srcp->loc_dcc_hook) {
		srcp->loc_dcc_hook(srcp, loc);
	} else {
		srcp_loc_dcc_put(loc);
	}

	return 0;
}

static int handler_ga_init(struct srcp *srcp, char *buffer)
{
	struct srcp_accessory_dcc *accessory;
	float rep_time;
	int bus;
	int address;
	char protocol[1024];
	int retcode;

	sscanf(buffer, "%f %d INFO %d GA %d "
	    "%s",
	    &rep_time, &retcode, &bus, &address,
	    protocol);

	accessory = calloc(sizeof(struct srcp_accessory_dcc), 1);
	if (!accessory)
		return -1;

	accessory->bus = bus;
	accessory->address = address;

	srcp_accessory_dcc_list_add(srcp, accessory);

	return 0;
}

static int handler_ga_get(struct srcp *srcp, char *readbuffer)
{
	struct srcp_accessory_dcc *accessory;
	float rep_time;
	int port, value, retcode, bus, address;

	sscanf(readbuffer, "%f %d INFO %d GA %d"
	    " %d %d",
	    &rep_time, &retcode, &bus, &address,
	    &port, &value);
			
	accessory = srcp_accessory_dcc_list_find(srcp, address);
	if (accessory && port < 2) {
		accessory->port[port] = value;
	
		if (srcp->accessory_dcc_hook) {
			srcp->accessory_dcc_hook(srcp, accessory);
		} else {
			srcp_accessory_dcc_put(accessory);
		}
	}

	return 0;
}

static int handler_gm(struct srcp *srcp, char *readbuffer)
{
	int send_to, reply_to;
	char msgtype[strlen(readbuffer)+1];
	char msg[strlen(readbuffer)+1];
	float rep_time;
	int retcode, bus;

	sscanf(readbuffer, "%f %d INFO %d GM"
	    " %d %d"
	    " %s %[^\n]",
	    &rep_time, &retcode, &bus,
	    &send_to, &reply_to,
	    msgtype, msg);
			
	if (srcp->message_hook) {
		srcp->message_hook(srcp, send_to, reply_to, msgtype, msg);
	}

	return 0;
}

static int handler_power_get(struct srcp *srcp, char *readbuffer)
{
	char state[4];
	float rep_time;
	int retcode, bus;
			
	sscanf(readbuffer, "%f %d INFO %d POWER %3s",
	    &rep_time, &retcode, &bus, state);
			
	if (!strcmp(state, "ON"))
		srcp->power = true;
	else
		srcp->power = false;
	if (srcp->power_hook) {
		srcp->power_hook(srcp, srcp->power);
	}

	return 0;
}

static int handler_lock_get(struct srcp *srcp, char *readbuffer)
{
	char group[4];
	float rep_time;
	int retcode, bus, address, duration, session_id;
	
	sscanf(readbuffer, "%f %d INFO %d LOCK %3s %d %d %d",
	    &rep_time, &retcode, &bus, group, &address, &duration, &session_id);

	if (!strcmp(group, "GL") ||
	    !strcmp(group, "SM")) {
		struct srcp_loc_dcc *loc;
		loc = srcp_loc_dcc_list_find(srcp, address);
		if (loc) {
			loc->lock = session_id;
			if (srcp->lock_loc_dcc_hook)
				srcp->lock_loc_dcc_hook(srcp, loc);
			else
				srcp_loc_dcc_put(loc);
		}
	} else if (!strcmp(group, "GA")) {
		struct srcp_accessory_dcc *accessory;
		accessory = srcp_accessory_dcc_list_find(srcp, address);
		if (accessory) {
			accessory->lock = session_id;
			if (srcp->lock_accessory_dcc_hook)
				srcp->lock_accessory_dcc_hook(srcp, accessory);
			else
				srcp_accessory_dcc_put(accessory);
		}
	} else {
		printf("Unknown group lock: %s\n", group);
	}
	
	return 0;
}

static int handler_lock_term(struct srcp *srcp, char *readbuffer)
{
	char group[4];
	float rep_time;
	int retcode, bus, address;
	
	sscanf(readbuffer, "%f %d INFO %d LOCK %3s %d",
	    &rep_time, &retcode, &bus, group, &address);

	if (!strcmp(group, "GL") ||
	    !strcmp(group, "SM")) {
		struct srcp_loc_dcc *loc;
		loc = srcp_loc_dcc_list_find(srcp, address);
		if (loc) {
			loc->lock = 0;
			if (srcp->lock_loc_dcc_hook)
				srcp->lock_loc_dcc_hook(srcp, loc);
			else
				srcp_loc_dcc_put(loc);
		}
	} else if (!strcmp(group, "GA")) {
		struct srcp_accessory_dcc *accessory;
		accessory = srcp_accessory_dcc_list_find(srcp, address);
		if (accessory) {
			accessory->lock = 0;
			if (srcp->lock_accessory_dcc_hook)
				srcp->lock_accessory_dcc_hook(srcp, accessory);
			else
				srcp_accessory_dcc_put(accessory);
		}
	} else {
		printf("Unknown group lock: %s\n", group);
	}
	
	return 0;
}


static int handler_nop(struct srcp *srcp, char *buffer)
{
	return 0;
}

struct srcp_response {
	int retcode;
	char *group;
	int (*handler)(struct srcp *srcp, char *buffer);
};

static struct srcp_response srcp_response_list[] = {
	{ SRCP_RESPONSE_INFO,      "GL",          handler_gl_get },
	{ SRCP_RESPONSE_INFO,      "GA",          handler_ga_get },
	{ SRCP_RESPONSE_INFO,      "GM",          handler_gm },
	{ SRCP_RESPONSE_INFO,      "POWER",       handler_power_get },
	{ SRCP_RESPONSE_INFO_INIT, "GL",          handler_gl_init },
	{ SRCP_RESPONSE_INFO_INIT, "GA",          handler_ga_init },
	{ SRCP_RESPONSE_INFO,      "LOCK",        handler_lock_get },
	{ SRCP_RESPONSE_INFO_TERM, "LOCK",        handler_lock_term },
	{ SRCP_RESPONSE_INFO_INIT, "POWER",       handler_nop },
	{ SRCP_RESPONSE_INFO,      "DESCRIPTION", handler_nop },
	{ SRCP_RESPONSE_INFO,      "SM",          handler_nop },
	{ SRCP_RESPONSE_INFO_INIT, "SM",          handler_nop },
	{ SRCP_RESPONSE_INFO_TERM, "SM",          handler_nop },
	{ SRCP_RESPONSE_INFO_INIT, "SESSION",     handler_nop },
	{ SRCP_RESPONSE_INFO_TERM, "SESSION",     handler_nop },
	{ SRCP_RESPONSE_INFO_INIT, "TIME",        handler_nop },
};

static int srcp_response_get(struct srcp *srcp)
{
	char readbuffer[1024];
	int retcode = -2;
	int i;
	
	if (fgets(readbuffer, sizeof(readbuffer), srcp->fdsock)) {
		float rep_time;
		int bus;
		char group[1024];

		if (srcp->all_hook)
			srcp->all_hook(srcp, readbuffer);
		
		sscanf(readbuffer, "%f %d INFO %d %s",
		    &rep_time, &retcode, &bus, group);
		
		if (retcode == SRCP_RESPONSE_OK)
			return retcode;

		if (SRCP_RESPONSE_ERROR(retcode))
			return retcode;

		for (i = 0; i < sizeof(srcp_response_list)/sizeof(struct srcp_response); i++) {
			if (retcode == srcp_response_list[i].retcode &&
			    !strcmp(group, srcp_response_list[i].group)) {
				int rh;
				
				rh = srcp_response_list[i].handler(srcp, readbuffer);
				
				if (rh)
					return -1;
				else 
					return retcode;
			}
		}

		printf("%s", readbuffer);
		printf("retcode: %d\n", retcode);
		printf("bus: %d\n", bus);
		printf("group: %s\n", group);

		printf("todo\n");
	} else {
		retcode = SRCP_RESPONSE_ERROR_NO_DATA;
	}

	return retcode;
}

int srcp_go(struct srcp *srcp)
{
	fprintf(srcp->fdsock, "GO\n");
	fflush(srcp->fdsock);
	
	return srcp_response_get_go(srcp);
}

int srcp_loc_dcc_set(struct srcp *srcp, struct srcp_loc_dcc *loc)
{
	int i;
	int nrofdecoderfunctions = loc->nrofdecoderfunctions;
	
	fprintf(srcp->fdsock, "SET %d GL %d %d %d %d",
	    loc->bus, loc->address, loc->drivemode, loc->V, loc->V_max);
	for (i = 0; i < nrofdecoderfunctions; i++) {
		fprintf(srcp->fdsock, " %d", loc->functions[i]);
	}
	fprintf(srcp->fdsock, "\n");
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_loc_dcc_get(struct srcp *srcp, struct srcp_loc_dcc *loc)
{
	fprintf(srcp->fdsock, "GET %d GL %d\n",
	    loc->bus, loc->address);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_accessory_dcc_set(struct srcp *srcp, struct srcp_accessory_dcc *acc, int port, bool value)
{
	fprintf(srcp->fdsock, "SET %d GA %d %d %d %d",
	    acc->bus, acc->address, port, value, -1);
	fprintf(srcp->fdsock, "\n");
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_accessory_dcc_get(struct srcp *srcp, struct srcp_accessory_dcc *acc)
{
	char readbuffer[100];
	int retcode;
	
	fprintf(srcp->fdsock, "GET %d GA %d 0\n",
	    acc->bus, acc->address);
	fflush(srcp->fdsock);
	if (fgets(readbuffer, sizeof(readbuffer), srcp->fdsock)) {
		float rep_time;
		int bus, address;
		int port, value;
		
		if (srcp->all_hook)
			srcp->all_hook(srcp, readbuffer);
		
		sscanf(readbuffer, "%f %d INFO %d GA %d"
		    " %d %d",
		    &rep_time, &retcode, &bus, &address,
		    &port, &value);
		if (retcode == 100) {
			acc->port[0] = value;
		}
	}
	fprintf(srcp->fdsock, "GET %d GA %d 1\n",
	    acc->bus, acc->address);
	fflush(srcp->fdsock);
	if (fgets(readbuffer, sizeof(readbuffer), srcp->fdsock)) {
		float rep_time;
		int bus, address;
		int port, value;
		
		if (srcp->all_hook)
			srcp->all_hook(srcp, readbuffer);
		
		sscanf(readbuffer, "%f %d INFO %d GA %d"
		    " %d %d",
		    &rep_time, &retcode, &bus, &address,
		    &port, &value);
		if (retcode == 100) {
			acc->port[1] = value;
		}
	}
	return retcode;
}

int srcp_info_handle(struct srcp *srcp)
{
	return srcp_response_get(srcp);
}

int srcp_loc_dcc_init(struct srcp *srcp, struct srcp_loc_dcc *loc)
{
	loc->refcount = 1;
	srcp_loc_dcc_list_add(srcp, loc);

	fprintf(srcp->fdsock, "INIT %d GL %d N %d %d %d\n",
	    loc->bus,
	    loc->address,
	    loc->protocolversion, 
	    loc->decoderspeedsteps,
	    loc->nrofdecoderfunctions);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_accessory_dcc_init(struct srcp *srcp, struct srcp_accessory_dcc *acc)
{
	fprintf(srcp->fdsock, "INIT %d GA %d N\n",
	    acc->bus,
	    acc->address);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_power_set(struct srcp *srcp, int bus, bool value)
{
	fprintf(srcp->fdsock, "SET %d POWER %s\n",
	    bus, value ? "ON" : "OFF");
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_power_get(struct srcp *srcp, int bus, bool *value)
{
	int ret;
	
	fprintf(srcp->fdsock, "SET %d POWER\n", bus);
	fflush(srcp->fdsock);

	ret = srcp_response_get(srcp);
	
	*value = srcp->power;
	
	return ret;
}

int srcp_hook_power(struct srcp *srcp, 
    void(*power_hook)(struct srcp *srcp, bool power))
{
	srcp->power_hook = power_hook;

	return 0;
}

int srcp_message_set(struct srcp *srcp, int send_to, int reply_to, char *type, char *msg)
{
	fprintf(srcp->fdsock, "SET 0 GM %d %d %s %s\n",
	    send_to, reply_to, type, msg);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_hook_message(struct srcp *srcp,
    void (*message_hook)(struct srcp *srcp, int send_to, int reply_to, char *type, char *msg))
{
	srcp->message_hook = message_hook;
	
	return 0;
}

int srcp_sm_init(struct srcp *srcp, int bus)
{
	fprintf(srcp->fdsock, "INIT %d SM NMRA\n", bus);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_sm_term(struct srcp *srcp, int bus)
{
	fprintf(srcp->fdsock, "TERM %d SM\n", bus);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_sm_set_cv(struct srcp *srcp, int bus, int address, int cv, int value)
{
	fprintf(srcp->fdsock, "SET %d SM %d CV %d %d\n", bus, address, cv, value);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

int srcp_lock_loc_dcc_set(struct srcp *srcp, struct srcp_loc_dcc *dcc)
{
	fprintf(srcp->fdsock, "SET %d LOCK GL %d 0\n",
	    dcc->bus,
	    dcc->address);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);

	return 0;
}

int srcp_lock_loc_dcc_term(struct srcp *srcp, struct srcp_loc_dcc *dcc)
{
	fprintf(srcp->fdsock, "TERM %d LOCK GL %d\n",
	    dcc->bus,
	    dcc->address);
	fflush(srcp->fdsock);

	return srcp_response_get(srcp);
}

