#define _BSD_SOURCE
#include <string.h>
#include <stdio.h>
#include <libudev.h>
#include <stdbool.h>

#include "udev_search.h"

char *udev_search_js2hidraw(char *joystick)
{
	struct udev *udev;
	struct udev_enumerate *enumerate;
	struct udev_list_entry *devices, *dev_list_entry;
	struct udev_device *dev;
	struct udev_device *parent = NULL, *hidparent;
	char *hidrawname = NULL;

	printf("joystick device: %s\n", joystick);

	/* Create the udev object */
	udev = udev_new();
	if (!udev) {
		printf("Can't create udev\n");
		return NULL;
	}
	
	enumerate = udev_enumerate_new(udev);
	udev_enumerate_add_match_subsystem(enumerate, "input");
	udev_enumerate_scan_devices(enumerate);
//	udev_enumerate_scan_subsystems(enumerate);

	devices = udev_enumerate_get_list_entry(enumerate);

	/* For each item enumerated, print out its information.
	   udev_list_entry_foreach is a macro which expands to
	   a loop. The loop will be executed for each member in
	   devices, setting dev_list_entry to a list entry
	   which contains the device's path in /sys. */
	udev_list_entry_foreach(dev_list_entry, devices) {
		const char *path, *node;
		struct udev_list_entry *devnode_list, *devnode;
		bool found = false;
		
		/* Get the filename of the /sys entry for the device
		   and create a udev_device object (dev) representing it */
		path = udev_list_entry_get_name(dev_list_entry);
		dev = udev_device_new_from_syspath(udev, path);

		/* usb_device_get_devnode() returns the path to the device node
		   itself in /dev. */
		node = udev_device_get_devnode(dev);
		if (node && !strcmp(joystick, node))
			found = true;

		devnode_list = udev_device_get_devlinks_list_entry(dev);
		udev_list_entry_foreach(devnode, devnode_list) {
			node = udev_list_entry_get_name(devnode);
			if (node && !strcmp(joystick, node))
				found = true;
		}

		/* The device pointed to by dev contains information about
		   the hidraw device. In order to get information about the
		   USB device, get the parent device with the
		   subsystem/devtype pair of "usb"/"usb_device". This will
		   be several levels up the tree, but the function will find
		   it.*/
		if (found) {
			parent = udev_device_get_parent_with_subsystem_devtype(
			       dev,
			       "usb",
			       "usb_device");
			if (!parent) {
				printf("Unable to find parent usb device.");
			}
	
			/* From here, we can call get_sysattr_value() for each file
			   in the device's /sys entry. The strings passed into these
			   functions (idProduct, idVendor, serial, etc.) correspond
			   directly to the files in the directory which represents
			   the USB device. Note that USB strings are Unicode, UCS2
			   encoded, but the strings returned from
			   udev_device_get_sysattr_value() are UTF-8 encoded. */
			printf("  VID/PID: %s %s\n",
			        udev_device_get_sysattr_value(parent,"idVendor"),
			        udev_device_get_sysattr_value(parent, "idProduct"));
			printf("  %s\n  %s\n",
			        udev_device_get_sysattr_value(parent,"manufacturer"),
			        udev_device_get_sysattr_value(parent,"product"));
			printf("  serial: %s\n",
			         udev_device_get_sysattr_value(parent, "serial"));
		} else {
			udev_device_unref(dev);
		}
	}
	/* Free the enumerator object */
	udev_enumerate_unref(enumerate);

	if (!parent) {
		return NULL;
	}

	enumerate = udev_enumerate_new(udev);
	udev_enumerate_add_match_subsystem(enumerate, "hidraw");
	udev_enumerate_scan_devices(enumerate);

	devices = udev_enumerate_get_list_entry(enumerate);

	udev_list_entry_foreach(dev_list_entry, devices) {
		const char *path, *node;
		
		path = udev_list_entry_get_name(dev_list_entry);
		dev = udev_device_new_from_syspath(udev, path);

		node = udev_device_get_devnode(dev);

		hidparent = udev_device_get_parent_with_subsystem_devtype(
		       dev,
		       "usb",
		       "usb_device");
		if (!hidparent) {
			printf("Unable to find parent usb device.");
		} else {
			if (!strcmp(
			    udev_device_get_sysattr_value(parent,"manufacturer"),
			    udev_device_get_sysattr_value(hidparent,"manufacturer")) &&
			    !strcmp(
			    udev_device_get_sysattr_value(parent,"product"),
			    udev_device_get_sysattr_value(hidparent,"product")) &&
			    !strcmp(
			    udev_device_get_sysattr_value(parent, "serial"),
			    udev_device_get_sysattr_value(hidparent, "serial"))) {
				printf("Found HIDRAW device: %s\n", node);
				hidrawname = strdup(node);
			}
		}
	
		udev_device_unref(dev);
	}
	/* Free the enumerator object */
	udev_enumerate_unref(enumerate);

	udev_unref(udev);

	return hidrawname;
}
